#ifndef id873715F57B504CCF8227CE03EA28CAFA
#define id873715F57B504CCF8227CE03EA28CAFA

#include "primitivetypes.hpp"
#include <boost/iterator/iterator_facade.hpp>
#include <iterator>
#include <vector>
#include <type_traits>

namespace dk {
	namespace implem {
		template <size_t D>
		size_t get_index_from_pos ( const Vector<CoordinateScalarType, D>& parPos, const Vector<CoordinateScalarType, D>& parSize ) a_pure;

		template <typename I, typename O>
		struct TypeWithQualifiers {
			typedef typename std::conditional<std::is_volatile<I>::value && std::is_const<I>::value,
						typename std::add_cv<O>::type,
						typename std::conditional<std::is_volatile<I>::value,
							typename std::add_volatile<typename std::remove_cv<O>::type>::type,
							typename std::conditional<std::is_const<I>::value,
								typename std::add_const<typename std::remove_cv<O>::type>::type,
								typename std::remove_cv<O>::type
							>::type
						>::type
					>::type result;
		};

#if defined(NDEBUG)
		template <>
		size_t get_index_from_pos<2> ( const Vector<CoordinateScalarType, 2>& parPos, const Vector<CoordinateScalarType, 2>& parSize ) a_pure;
#endif

		template <size_t D>
		Vector<CoordinateScalarType, D> buildPastEndCoordinate ( const Vector<CoordinateScalarType, D>& parFrom, const Vector<CoordinateScalarType, D>& parTo ) a_pure;
	} //namespace implem

	template <typename T, size_t D, typename T1=typename std::remove_cv<T>::type>
	class TileIterator : public boost::iterator_facade<TileIterator<T, D>, T, boost::bidirectional_traversal_tag> {
		friend class boost::iterator_core_access;
		typedef std::vector<T1> vector_type;
		typedef typename implem::TypeWithQualifiers<T, vector_type>::result qualif_vector_type;
	public:
		typedef Vector<CoordinateScalarType, D> coords;

		TileIterator ( void );
		TileIterator ( const TileIterator& parOther ) = default;
		TileIterator ( TileIterator&& parOther ) = default;
		TileIterator ( qualif_vector_type* parData, const coords& parFrom, const coords& parTo );
		TileIterator ( qualif_vector_type* parData, const coords& parFrom, const coords& parTo, const coords& parAreaFrom, const coords& parAreaTo );
		~TileIterator ( void ) = default;

		const coords& position ( void ) const { return m_pos; }

	private:
		void increment ( void );
		void decrement ( void );
		void advance ( size_t parAdvance );
		ptrdiff_t distance_to ( const TileIterator& parOther );
		bool equal ( const TileIterator& parOther ) const;
		T& dereference ( void ) const { return (*m_data)[get_current_index()]; }
		size_t get_current_index ( void ) const { return implem::get_index_from_pos<D>(m_pos, m_areato - m_areafrom); }

		coords m_pos;
		coords m_from;
		coords m_to;
		coords m_areafrom;
		coords m_areato;
		qualif_vector_type* m_data;
	};
} //namespace dk

#include "implem/tileiterator.inl"

#endif
